#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2024 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# @author Paul Dechamps
# @date 2024
# Optimization for a transonic case

from fwk.wutils import parseargs
import fwk
from fwk.testing import *
from fwk.coloring import ccolors

from blast.api.mda_api import BlasterSolver
from blast.api.mda_api import BlasterModeUpdater

import numpy as np
import openmdao.api as om

def cfgInviscid(nthrds, verb):
    import os.path
    # Parameters
    return {
    # Options
    'scenario' : 'aerodynamic',
    'task' : 'optimization',
    'Threads' : nthrds, # number of threads
    'Verb' : verb, # verbosity
    # Model (geometry or mesh)
    'File' : os.path.dirname(os.path.dirname(os.path.abspath(__file__))) + '/models/dart/n0012.geo', # Input file containing the model
    'Pars' : {'xLgt' : 50, 'yLgt' : 50, 'msF' : 20, 'msTe' : 0.01, 'msLe' : 0.008}, # parameters for input file model
    'Dim' : 2, # problem dimension
    'Format' : 'gmsh', # save format (vtk or gmsh)
    # Markers
    'Fluid' : 'field', # name of physical group containing the fluid
    'Farfield' : ['upstream', 'farfield', 'downstream'], # LIST of names of physical groups containing the farfield boundaries (upstream/downstream should be first/last element)
    'Wing' : 'airfoil', # LIST of names of physical groups containing the lifting surface boundary
    'Wake' : 'wake', # LIST of names of physical group containing the wake
    'WakeTip' : 'wakeTip', # LIST of names of physical group containing the edge of the wake
    'Te' : 'te', # LIST of names of physical group containing the trailing edge
    'Upstream' : 'upstream',
    # Freestream
    'M_inf' : 0.78, # freestream Mach number
    'AoA' : 0., # freestream angle of attack
    # Geometry
    'S_ref' : 1., # reference surface length
    'c_ref' : 1., # reference chord length
    'x_ref' : .25, # reference point for moment computation (x)
    'y_ref' : 0.0, # reference point for moment computation (y)
    'z_ref' : 0.0, # reference point for moment computation (z)
    # Numerical
    'LSolver' : 'SparseLu', # inner solver (Pardiso, MUMPS or GMRES)
    'G_fill' : 2, # fill-in factor for GMRES preconditioner
    'G_tol' : 1e-5, # tolerance for GMRES
    'G_restart' : 50, # restart for GMRES
    'Rel_tol' : 1e-6, # relative tolerance on solver residual
    'Abs_tol' : 1e-8, # absolute tolerance on solver residual
    'Max_it' : 50 # solver maximum number of iterations
    }

def cfgBlast(verb):
    return {
        'Re' : 1e8,                 # Freestream Reynolds number
        'Minf' : 0.78,               # Freestream Mach number (used for the computation of the time step only)
        'CFL0' : 1.,                # Inital CFL number of the calculation
        'Verb': verb,               # Verbosity level of the solver
        'couplIter': 75,            # Maximum number of iterations
        'couplTol' : 1e-4,          # Tolerance of the VII methodology
        'iterPrint': 5,             # int, number of iterations between outputs
        'resetInv' : True,          # bool, flag to reset the inviscid calculation at every iteration.
        'sections' : [0],           # List of sections for boundary layer calculation
        'xtrF' : [0.05, 0.05],       # Forced transition location
        'interpolator' : 'Matching',# Interpolator for the coupling
        'nSections' : 1,            # Number of sections in the domain
        'span' : 1.0                # Wing span
    }

class groupMDA(om.Group):
    def setup(self):
        args = parseargs()
        icfg = cfgInviscid(args.k, args.verb)
        vcfg = cfgBlast(args.verb)
        gr = 1.1
        if gr == 1.0:
            icfg['Pars']['msF'] = icfg['Pars']['msLe']
        else:
            n = np.log10(1-(1-gr)*icfg['Pars']['xLgt']/icfg['Pars']['msLe'])/np.log10(gr)
            icfg['Pars']['msF'] = icfg['Pars']['msLe']*gr**(n-1)

        # Add the MDAO components
        blaster = BlasterSolver(vcfg, icfg, iSolverName='DART', saveAll=False)

        self.add_subsystem('dvs', om.IndepVarComp(), promotes=['*']) # Design variables box
        self.add_subsystem('mod', BlasterModeUpdater(tolerance=10., force_mode='visc')) # Mode updater
        self.add_subsystem('mesh', blaster.getBoundaryMesh()) # Initial mesh
        self.add_subsystem('geometry', blaster.getBoundaryFFDGeometry()) # Geometry computation
        # Add an intermediate variable for the constraint (such that upper_shape[0] = lower_shape[0])
        self.add_subsystem('constraint_comp', om.ExecComp('constraint_var = lower_shape[0] + upper_shape[0]',
                                                          lower_shape=np.zeros(4), upper_shape=np.zeros(4)), promotes=['*'])
        self.add_subsystem('solver', blaster) # Blaster solver
        self.add_subsystem('obj_cmp', om.ExecComp('obj = cd'), promotes=['obj', 'cd']) # Objective function

        # Connect meshes
        self.connect('mesh.x_aero0', 'geometry.x_aero0')
        self.connect('geometry.x_aero_out', 'solver.x_aero')
        # Connect mode updater
        self.connect('mod.visc_mode', 'solver.visc_mode')

        #nonlinear_solver = om.NewtonSolver(solve_subsystems=False)
        self.linear_solver = om.DirectSolver()
    
    def configure(self):
        # Design vars
        self.dvs.add_output('aoa', val=0.)
        self.connect('aoa', 'solver.aoa')
        self.add_design_var('aoa', lower=-5, upper=5)

        self.dvs.add_output('upper_shape', val=[0.17553542, 0.14723497, 0.14531762, 0.13826955]) # NACA 0012
        #self.dvs.add_output('upper_shape', val=[0.25367349, 0.24936591, 0.26996167, 0.28365274]) # NACA4412
        self.connect('upper_shape', 'geometry.upper_shape')
        self.add_design_var('upper_shape', lower=-0.15, upper=0.25, scaler=10)

        self.dvs.add_output('lower_shape', val=[-0.16935506, -0.15405552, -0.13912672, -0.14168947]) # NACA 0012
        #self.dvs.add_output('lower_shape', val=[-0.1330126,  -0.01246078, -0.04065711, -0.0123221]) # NACA 4412
        self.connect('lower_shape', 'geometry.lower_shape')
        self.add_design_var('lower_shape', lower=-0.2, upper=0.2, scaler=10)

        self.dvs.add_output('class_shape_n1', val=0.5)
        self.connect('class_shape_n1', 'geometry.class_shape_n1')
        # self.add_design_var('class_shape_n1', lower=-0.5, upper=0.1)

        self.dvs.add_output('class_shape_n2', val=1.)
        self.connect('class_shape_n2', 'geometry.class_shape_n2')
        # self.add_design_var('class_shape_n2', lower=-0.5, upper=0.1)

        # Constraints
        self.add_constraint('solver.cl', equals=0.4, scaler=10)
        self.add_constraint('geometry.thickness', lower=0.12, linear=True, scaler=10)
        self.add_constraint('geometry.LEradius', equals=0.0224, linear=True, scaler=100)
        self.add_constraint('constraint_var', equals=0., linear=True)

        self.connect('solver.cd', 'cd')
        self.add_objective('obj', scaler=1000)

def main():
    # Timer.
    tms = fwk.Timers()
    tms['total'].start()

    prob = om.Problem()
    prob.model = groupMDA()

    prob.driver = om.ScipyOptimizeDriver()
    prob.driver.options['optimizer'] = 'SLSQP'
    prob.driver.options['tol'] = 1e-4

    recorder = om.SqliteRecorder('reports/BlasterSolver_file.sql')

    prob.driver.add_recorder(recorder)
    prob.driver.recording_options['record_objectives'] = True
    prob.driver.recording_options['record_desvars'] = True
    prob.driver.recording_options['includes'] = ['aoa', 'solver.cl', 'solver.cd', 'mesh.x_aero0', 'geometry.x_aero_out']

    prob.setup()
    # print('check')
    # prob.check_partials(includes=['cl', 'cd'], compact_print=True)
    # print('done')
    prob.run_driver()

    # problem timers
    print(prob.model.solver.tms)

    # Case reader
    cr = om.CaseReader(prob.get_reports_dir() / "../BlasterSolver_file.sql")
    driver_cases = cr.list_cases('driver', out_stream=None)

    # Get evolution of f and x
    aoa_evol = []
    cl_evol = []
    cd_evol = []
    shape = []
    shape0 = []

    for c in driver_cases:
        case = cr.get_case(c)
        aoa_evol.append(case.get_val('aoa')[0])
        cl_evol.append(case.get_val('solver.cl')[0])
        cd_evol.append(case.get_val('solver.cd')[0])
        shape0.append(case.get_val('mesh.x_aero0'))
        shape.append(case.get_val('geometry.x_aero_out'))

    print('aoa:', prob['aoa'])
    print('cl:', prob['solver.cl'])
    print('cd:', prob['solver.cd'])
    print('obj:', prob['obj'])

    # eof
    tms['total'].stop()
    print(tms)

if __name__ == "__main__":
    main()
